#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os

from zenml.utils.version import __version__


def handle_bool_env_var(var, default=False):
    """Converts normal env var to boolean"""
    var = os.getenv(var, None)
    for i in ['1', 'y', 'yes', 'True', 'true']:
        if i == var:
            return True
    return default


# Global constants
APP_NAME = 'zenml'
CONFIG_VERSION = '1'

# Environment variables
ZENML_PIPELINE_CONFIG = 'ZENML_PIPELINE_CONFIG'
CE_CLOUD_JOB_PREFIX = 'CE_CLOUD_JOB_PREFIX'
ENV_ZENML_DEBUG = 'ZENML_DEBUG'

# Logging variables
ENV_ZENML_LOGGING_VERBOSITY = 'ZENML_LOGGING_VERBOSITY'
# TODO: [HIGH] Switch to default False before release
IS_DEBUG_ENV = handle_bool_env_var(ENV_ZENML_DEBUG, default=True)
if IS_DEBUG_ENV:
    ZENML_LOGGING_VERBOSITY = os.getenv(ENV_ZENML_LOGGING_VERBOSITY, 4)
else:
    ZENML_LOGGING_VERBOSITY = os.getenv(ENV_ZENML_LOGGING_VERBOSITY, 0)

# Base images for zenml
ZENML_REGISTRY = 'eu.gcr.io/maiot-zenml/'
ZENML_BASE_IMAGE_NAME = f'{ZENML_REGISTRY}/zenml:base-{__version__}'
ZENML_TRAINER_IMAGE_NAME = f'{ZENML_REGISTRY}/zenml:cuda-{__version__}'

# Evaluation utils constants
COMPARISON_NOTEBOOK = 'comparison_notebook.ipynb'
EVALUATION_NOTEBOOK = 'evaluation_notebook.ipynb'

# Pipeline related constants
PREPROCESSING_FN = \
    'zenml.core.components.transform.transform_module.preprocessing_fn'
TRAINER_FN = 'zenml.core.components.trainer.trainer_module.run_fn'
EVALUATOR_MODULE_FN = 'zenml.core.components.evaluator.evaluator_module'



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Here lies the logic of running the pipeline. It consumes the input as
configuration"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json
import time
import uuid
from typing import Dict, Union
from typing import Text

import yaml


class PipelineRunner:
    """PipelineRunner runs (TFX) pipelines with appropriate backends.

    All backends need to get plugged in this class.
    """
    def __init__(self):
        pass

    def run(self, environment_yaml: Union[Text, Dict],
            experiment_yaml: Union[Text, Dict]):
        # Create the required gzipped tar file
        prepare_sdist()

        if isinstance(environment_yaml, Text) \
                and isinstance(experiment_yaml, Text):
            # Read both config files
            environment_dict = yaml.load(open(environment_yaml))
            experiment_dict = yaml.load(open(experiment_yaml))
        elif isinstance(environment_yaml, Dict) \
                and isinstance(experiment_yaml, Dict):
            environment_dict = environment_yaml
            experiment_dict = experiment_yaml
        else:
            raise AssertionError(
                'Params have to either both dict or both paths.')

        # Create a dict which includes the environment settings of the pipeline
        env_spec = get_pipeline_spec(environment_dict)
        # Create a proper JSON file for the experiment config
        config = create_experiment_json(experiment_dict, env_spec)
        str_config = json.dumps(config)

        # Create and execute the pipeline
        created_pipeline = create_normal_pipeline(
            env_spec,
            experiment_args=experiment_dict,
            cloud_job_prefix=str(time.time())
        )

        orchestrator = env_spec['orchestrator']

        if orchestrator == 'kubeflow':

            from tfx.orchestration.kubeflow import kubeflow_dag_runner
            from tfx.orchestration.kubeflow.kubeflow_dag_runner import \
                KubeflowDagRunner
            from gdp.utils.environment_utils import add_ce_env

            pipeline_operator_funcs = \
                kubeflow_dag_runner.get_default_pipeline_operator_funcs()
            pipeline_operator_funcs.append(
                add_ce_env(ZENML_PIPELINE_CONFIG, str_config))

            runner_config = kubeflow_dag_runner.KubeflowDagRunnerConfig(
                kubeflow_metadata_config=env_spec[
                    'metadata_connection_config'],
                tfx_image=GDP_IMAGE,
                pipeline_operator_funcs=pipeline_operator_funcs,
            )
            runner = KubeflowDagRunner(config=runner_config)
            runner._pipeline_id = str(uuid.uuid4())

            def pipeline_func():
                runner._construct_pipeline_graph(created_pipeline,
                                                 env_spec['pipeline_root'])

            # import kfp
            # from datetime import datetime
            # kfp_client = kfp.Client(
            #     host='https://kf-dev-04.endpoints.core-engine.cloud.goog
            #     /pipeline',
            #     client_id='973445798975-1brrhd1bmalcggu7v6aro8ttohmlpj8l.apps
            #     .googleusercontent.com',
            #     namespace='hamza',
            # )
            # test = kfp_client.create_run_from_pipeline_func(
            #     pipeline_func=pipeline_func,
            #     arguments={},
            #     run_name='remote_{}'.format(str(datetime.now())),
            #     experiment_name='maiot-version-upgrade-0214',
            # )
            return pipeline_func

        elif orchestrator == 'beam':
            from tfx.orchestration.beam.beam_dag_runner import BeamDagRunner
            import os
            # for local runs we add the env variable here
            os.environ[ZENML_PIPELINE_CONFIG] = str_config
            runner = BeamDagRunner(env_spec['orchestration_args'])
            runner.run(created_pipeline)
        elif orchestrator == 'ce':
            from tfx.orchestration.beam.beam_dag_runner import BeamDagRunner
            import os
            # for local runs we add the env variable here
            os.environ[ZENML_PIPELINE_CONFIG] = str_config
            runner = CERunner(env_spec[GlobalKeys.API_ARGS_][APIKeys.TOKEN_],
                              env_spec[GlobalKeys.API_ARGS_][APIKeys.HOST_],
                              env_spec[GlobalKeys.API_ARGS_][
                                  APIKeys.ENDPOINT_],
                              env_spec['orchestration_args'])
            runner.run(created_pipeline)
        else:
            raise Exception('What the hell is wrong with you?')


if __name__ == "__main__":
    # TODO: [LOW] Write a way to run this independantly.
    pass



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import datetime
import os
import re
import shutil
from distutils.core import run_setup

from tfx.orchestration import metadata
from tfx.orchestration.kubeflow.proto import kubeflow_pb2
from tfx.utils import io_utils
from zenml.utils.standards import standard_environment as std_env

from zenml.utils.constants import TrainingTypes


def prepare_sdist():
    """Refer to the README.md in the docs folder"""
    dist_path = os.path.join(os.getcwd(), 'dist')

    if os.path.exists(dist_path) and os.path.isdir(dist_path):
        print('Removing {}'.format(dist_path))
        shutil.rmtree(dist_path)
    else:
        print('There is no dist folder.')

    # FOR THE BOPA
    run_setup('setup.py', script_args=['sdist'])
    req_path = os.path.join(dist_path, 'requirements.txt')
    io_utils.write_string_file(req_path, '')

    # FOR THE BEPA
    run_setup('setup.py', script_args=['sdist'])
    req_path = os.path.join(dist_path, 'requirements.txt')
    io_utils.write_string_file(req_path, '')


def create_experiment_json(experiment_dict, env_spec):
    # Add the bq args which resides in the env dict to the exp dict
    # We have to do this because SplitGen gets bq_args from exp_config for now
    """
    Args:
        experiment_dict:
        env_spec:
    """
    experiment_dict['bq_args'] = env_spec['bq_args']

    ### OUTDATED: TO BE REVISITED SOMETIME ###
    # Save the resulting JSON config file inside the GDP package
    # import json
    # gdp_path = os.path.join(os.getcwd(),
    #                         PACKAGE_NAME,
    #                         'experiment.config.json')
    # with open(gdp_path, 'w') as f:
    #     json.dump(experiment_dict, f)
    return experiment_dict


def create_metadata_connection_config(metadata_args,
                                      pipeline_name,
                                      output_base_dir):
    """Converts yaml metadata args to a ML Metadata config instance according to
    type specified in the yaml. For now, sqlite and mysql are supported.

    Args:
        metadata_args: dict specified in the config.yml
        pipeline_name: str, name of the pipeline
        output_base_dir: str, the path to the output directory

    Returns:
        instance of type metadata_store_pb2.ConnectionConfig
    """
    if metadata_args[std_env.MetadataKeys.TYPE] == 'mysql':
        return metadata.mysql_metadata_connection_config(
            host=metadata_args[std_env.MetadataKeys.HOST_],
            port=metadata_args[std_env.MetadataKeys.PORT_],
            database=metadata_args[std_env.MetadataKeys.DATABASE_],
            username=metadata_args[std_env.MetadataKeys.USERNAME_],
            password=metadata_args[std_env.MetadataKeys.PASSWORD_]
        )
    elif metadata_args[std_env.MetadataKeys.TYPE] == 'kubeflow_mysql':
        # metadata_config = \
        #     kubeflow_dag_runner.get_default_kubeflow_metadata_config()

        metadata_config = kubeflow_pb2.KubeflowMetadataConfig()
        metadata_config.mysql_db_service_host.value = metadata_args[
            std_env.MetadataKeys.HOST_]
        metadata_config.mysql_db_service_port.value = str(
            metadata_args[std_env.MetadataKeys.PORT_])
        metadata_config.mysql_db_name.value = metadata_args[
            std_env.MetadataKeys.DATABASE_]
        metadata_config.mysql_db_user.value = metadata_args[
            std_env.MetadataKeys.USERNAME_]
        metadata_config.mysql_db_password.value = metadata_args[
            std_env.MetadataKeys.PASSWORD_]
        return metadata_config
    elif metadata_args[std_env.MetadataKeys.TYPE] == 'sqlite':
        db_path = os.path.join(output_base_dir,
                               'database',
                               pipeline_name,
                               'metadata.db')
        return metadata.sqlite_metadata_connection_config(db_path)
    else:
        raise AttributeError('Unsupported metadata db type. Choose one of '
                             '"mysql" or "sqlite"')


def parse_yaml_beam_args(pipeline_args):
    """Converts yaml beam args to list of args TFX accepts

    Args:
        pipeline_args: dict specified in the config.yml

    Returns:
        list of strings, where each string is a beam argument
    """
    return ['--{}={}'.format(key, value) for key, value in
            pipeline_args.items()]


def generate_unique_name(base_name):
    """
    Args:
        base_name:
    """
    identifier = os.getenv('CI_COMMIT_SHORT_SHA', os.getenv('USER', 'local'))
    return re.sub(
        r'[^0-9a-zA-Z-]+',
        '-',
        '{pipeline_name}-{identifier}-{ts}'.format(
            pipeline_name=base_name,
            identifier=identifier,
            ts=int(datetime.datetime.timestamp(datetime.datetime.now()))
        ).lower()
    )


def add_ce_env(env_var_name: str, env_var_value: str):
    """A function thats lets you inject environment variable inside a pod in
    Kubeflow

    Args:
        env_var_name (str):
        env_var_value (str):
    """
    import kfp.dsl as dsl
    from kubernetes.client.models import V1EnvVar

    def m(container_op: dsl.ContainerOp):
        env_var = V1EnvVar(name=env_var_name, value=env_var_value)
        container_op.container.add_env_variable(env_var)

    return m


def check_config(env_dict):
    """
    Args:
        env_dict:
    """
    std_env.GlobalKeys.key_check(env_dict)

    std_env.BigQueryKeys.key_check(env_dict[std_env.GlobalKeys.BQ_ARGS])

    std_env.OrchestratorKeys.key_check(
        env_dict[std_env.GlobalKeys.ORCHESTRATION])
    orchestrator = env_dict[std_env.GlobalKeys.ORCHESTRATION][
        std_env.OrchestratorKeys.TYPE]
    orchestrator_args = env_dict[std_env.GlobalKeys.ORCHESTRATION][
        std_env.OrchestratorKeys.ARGS]
    if orchestrator == 'beam':
        std_env.BeamKeys.key_check(orchestrator_args)
    elif orchestrator == 'kubeflow':
        std_env.KubeflowKeys.key_check(orchestrator_args)
    elif orchestrator == 'ce':
        pass
    else:
        raise AssertionError('Unknown type for the orchestration!')

    std_env.TrainingKeys.key_check(env_dict[std_env.GlobalKeys.TRAINING])
    training_type = env_dict[std_env.GlobalKeys.TRAINING][
        std_env.TrainingKeys.TYPE]
    training_args = env_dict[std_env.GlobalKeys.TRAINING][
        std_env.TrainingKeys.ARGS]
    if training_type == TrainingTypes.gcaip.name:
        std_env.AIPlatformKeys.key_check(training_args)
        tiers = ['BASIC', 'STANDARD_1', 'PREMIUM_1', 'BASIC_GPU',
                 'BASIC_TPU', 'CUSTOM']
        assert training_args[std_env.AIPlatformKeys.SCALE_TIER] in tiers, \
            'Invalid scale tier. Select one from {}'.format(tiers)
    elif training_type == TrainingTypes.local.name:
        pass
    else:
        raise AssertionError('Unknown type {} for the training! Choose either'
                             'local or gcaip.'.format(training_type))

    std_env.BeamKeys.key_check(env_dict[std_env.GlobalKeys.EXECUTION])

    std_env.MetadataKeys.key_check(env_dict[std_env.GlobalKeys.METADATA_ARGS])



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import click

from zenml.cli.cli import cli, pass_config


@cli.group()
def analytics():
    """Analytics for opt-in and opt-out"""
    pass


@analytics.command('opt-in',
                   context_settings=dict(ignore_unknown_options=True))
@pass_config
def opt_in(info):
    """Opt-in to analytics"""
    info.set_analytics_opt_in(True)
    click.echo('Opted in to analytics.')


@analytics.command('opt-out',
                   context_settings=dict(ignore_unknown_options=True))
@pass_config
def opt_out(info):
    """Opt-out of analytics"""
    info.set_analytics_opt_in(False)
    click.echo('Opted out of analytics.')



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Factory to register step classes to steps"""

from zenml.core.steps.deployer import DeployerStep
from zenml.core.steps.evaluator import EvaluatorStep
from zenml.core.steps.preprocesser import PreprocesserStep
from zenml.core.steps.split import SplitStep
from zenml.core.steps.timeseries import TimeseriesStep
from zenml.core.steps.trainer import TrainerStep


class StepFactory:
    """Definition of StepFactory to track all steps in ZenML.

    All steps (including custom steps) are to be registered here.
    """

    def __init__(self):
        self.steps = {}

    def get_steps(self):
        return self.steps

    def get_single_step(self, key):
        return self.steps[key]

    def register_step(self, key, step_):
        self.steps[key] = step_


# Register the injections into the factory
type_factory = StepFactory()
type_factory.register_step('trainer', TrainerStep)
type_factory.register_step('split', SplitStep)
type_factory.register_step('evaluator', EvaluatorStep)
type_factory.register_step('timeseries', TimeseriesStep)
type_factory.register_step('preprocessor', PreprocesserStep)
type_factory.register_step('deployer', DeployerStep)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import base64
import os
from typing import Text

import click
import nbformat as nbf

from zenml.utils.constants import APP_NAME, EVALUATION_NOTEBOOK, \
    COMPARISON_NOTEBOOK
from zenml.utils.evaluation_utils import get_eval_block, get_model_block, \
    import_block, info_block, application_block, interface_block


def view_statistics(artifact_uri, magic: bool = False):
    """
    Args:
        artifact_uri (Text):
        magic (bool):
    """
    import os
    import tensorflow as tf
    from tensorflow_metadata.proto.v0 import statistics_pb2
    import panel as pn

    result = {}
    for split in os.listdir(artifact_uri):
        stats_path = os.path.join(artifact_uri, split, 'stats_tfrecord')
        serialized_stats = next(
            tf.compat.v1.io.tf_record_iterator(stats_path))
        stats = statistics_pb2.DatasetFeatureStatisticsList()
        stats.ParseFromString(serialized_stats)
        dataset_list = statistics_pb2.DatasetFeatureStatisticsList()
        for i, d in enumerate(stats.datasets):
            d.name = split
            dataset_list.datasets.append(d)
        result[split] = dataset_list
    h = get_statistics_html(result)

    if magic:
        import sys
        if 'ipykernel' not in sys.modules:
            raise EnvironmentError(
                'The magic functions are only usable '
                'in a Jupyter notebook.')
        from IPython.core.display import display, HTML
        display(HTML(h))

    else:
        pn.serve(panels=pn.pane.HTML(h, width=1200), show=True)


def get_statistics_html(stats_dict):
    """
    Args:
        stats_dict:
    """
    from tensorflow_metadata.proto.v0 import statistics_pb2

    combined_statistics = statistics_pb2.DatasetFeatureStatisticsList()

    for split, stats in stats_dict.items():
        stats_copy = combined_statistics.datasets.add()
        stats_copy.MergeFrom(stats.datasets[0])
        stats_copy.name = split

    protostr = base64.b64encode(
        combined_statistics.SerializeToString()).decode('utf-8')

    html_template = """<iframe id='facets-iframe' width="100%" height="500px"></iframe>
        <script>
        facets_iframe = document.getElementById('facets-iframe');
        facets_html = '<script src="https://cdnjs.cloudflare.com/ajax/libs/webcomponentsjs/1.3.3/webcomponents-lite.js"><\/script><link rel="import" href="https://raw.githubusercontent.com/PAIR-code/facets/master/facets-dist/facets-jupyter.html"><facets-overview proto-input="protostr"></facets-overview>';
        facets_iframe.srcdoc = facets_html;
         facets_iframe.id = "";
         setTimeout(() => {
           facets_iframe.setAttribute('height', facets_iframe.contentWindow.document.body.offsetHeight + 'px')
         }, 1500)
         </script>"""

    # pylint: enable=line-too-long
    html = html_template.replace('protostr', protostr)
    return html


def get_schema_proto(artifact_uri: Text):
    from google.protobuf import text_format
    from tensorflow.python.lib.io import file_io
    from tensorflow_metadata.proto.v0 import schema_pb2

    schema_path = os.path.join(artifact_uri, 'schema.pbtxt')
    schema = schema_pb2.Schema()
    contents = file_io.read_file_to_string(schema_path)
    schema = text_format.Parse(contents, schema)
    return schema


def evaluate_single_pipeline(trainer_path: Text, eval_path: Text,
                             magic: bool = False):
    # Resolve the pipeline run_id
    """
    Args:
        pipeline_name (Text):
        magic (bool):
    """

    # Patch to make it work locally
    import json
    with open(os.path.join(eval_path, 'eval_config.json'), 'r') as f:
        eval_config = json.load(f)
    eval_config['modelLocations'][''] = eval_path
    with open(os.path.join(eval_path, 'eval_config.json'), 'w') as f:
        json.dump(eval_config, f)

    if magic:
        from zenml.utils.shell_utils import create_new_cell
        model_block = get_model_block(trainer_path)
        eval_block = get_eval_block(eval_path)

        create_new_cell(eval_block)
        create_new_cell(model_block)

    else:
        nb = nbf.v4.new_notebook()
        nb['cells'] = [
            nbf.v4.new_code_cell(get_model_block(trainer_path)),
            nbf.v4.new_code_cell(get_eval_block(eval_path))]

        config_folder = click.get_app_dir(APP_NAME)

        if not (os.path.exists(config_folder) and os.path.isdir(
                config_folder)):
            os.makedirs(config_folder)

        final_out_path = os.path.join(config_folder, EVALUATION_NOTEBOOK)
        s = nbf.writes(nb)
        if isinstance(s, bytes):
            s = s.decode('utf8')

        with open(final_out_path, 'w') as f:
            f.write(s)
        os.system('jupyter notebook "{}"'.format(final_out_path))


def compare_multiple_pipelines():
    """
    Args:
        workspace_id (Text):
    """
    info = {}

    # generate notebook
    nb = nbf.v4.new_notebook()
    nb['cells'] = [
        nbf.v4.new_code_cell(import_block()),
        nbf.v4.new_code_cell(info_block(info)),
        nbf.v4.new_code_cell(application_block()),
        nbf.v4.new_code_cell(interface_block()),
    ]

    # write notebook
    config_folder = click.get_app_dir(APP_NAME)

    if not (os.path.exists(config_folder) and os.path.isdir(
            config_folder)):
        os.makedirs(config_folder)

    final_out_path = os.path.join(config_folder, COMPARISON_NOTEBOOK)
    s = nbf.writes(nb)
    if isinstance(s, bytes):
        s = s.decode('utf8')

    with open(final_out_path, 'w') as f:
        f.write(s)

    # serve notebook
    os.system('panel serve "{}" --show'.format(final_out_path))
